/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package main

import (
	"context"
	"errors"
	"fmt"
	"io"
	"log"
	"os"
	"os/signal"
	"time"

	"golang.org/x/sys/unix"
)

// -3 silent
// -2 fatal
// -1 error
//
//	0 warn
//	1 info
//	2 debug
var LogLevel int

func debug(fmt string, msg ...any) {
	if LogLevel < 2 {
		return
	}
	fmt = "debug: " + fmt
	if len(msg) == 0 {
		log.Printf(fmt)
		return
	}
	log.Printf(fmt, msg...)
}

func main() {
	ctx, cancel := context.WithCancel(context.Background())

	go func() {
		sigChan := make(chan os.Signal, 1) // catch signals
		signal.Notify(
			sigChan,
			unix.SIGINT,
			unix.SIGTERM,
		)
		for {
			switch sig := <-sigChan; sig {
			case unix.SIGINT, unix.SIGTERM:
				log.Printf("Caught SIGINT/SIGTERM.")
				cancel()
			}
		}
	}()

	if len(os.Args) != 2 {
		log.Fatalf("Usage: %s <log_file>", os.Args[0])
	}
	logPath := os.Args[1]

	os.Exit(mainLoop(ctx, logPath))
}

func mainLoop(ctx context.Context, path string) int {
	var file *os.File
	var lineCnt uint64

	inode, size, err := reopenFile(path, &file)
	if err != nil {
		log.Printf("Could not open file at %#v: %v", path, err)
		return 1
	}
	if file == nil {
		panic("BUG: Received nil file pointer")
	}
	initBuf := make([]byte, 0, L1CacheSize)
	buf := initBuf

	for {
		if ctx.Err() != nil {
			return 2
		}

		i := 0
		for i < len(buf) {
			if buf[i] == '\n' {
				break
			}
			i++
		}

		if i != len(buf) {
			line := buf[:i]
			debug("Detected linefeed; processing line %#v", string(line))
			if err = processLine(line); err != nil {
				log.Printf("Error while processing line: %v", err)
			}
			buf = buf[i+1:]
			lineCnt++
			continue
		}

		if len(buf) == cap(buf) {
			debug("Reached end of buffer; resetting cursor to start of buffer")
			nextBuf := initBuf[0:len(buf)]
			copy(nextBuf, buf)
			buf = nextBuf
		}

		n, err := file.Read(buf[len(buf):cap(buf)])
		if errors.Is(err, io.EOF) {
			var rot bool
			rot, err = fileWasRotatedOrTruncated(path, inode, size)
			if rot {
				debug("File %#v was rotated or truncated; reopening...", path)
				inode, size, err = reopenFile(path, &file)
				if err != nil {
					log.Printf("Could not open file at %#v: %v", path, err)
					return 1
				}
				continue
			}
			time.Sleep(100 * time.Millisecond)
			continue
		}
		if err != nil {
			log.Printf("Unexpected read error: %v", err)
			return 3
		}
		buf = buf[:len(buf)+n]
	}
}

func processLine(line []byte) error {
	fmt.Println(string(line))
	return nil
}

func reopenFile(
	path string,
	file **os.File,
) (inode uint64, size int64, err error) {
	_ = (*file).Close()
	f, err := os.Open(path)
	if err != nil {
		err = fmt.Errorf("Could not open file at %#v: %w", path, err)
		return
	}
	(*file) = f
	stat := &unix.Stat_t{}
	if err = unix.Stat(path, stat); err != nil {
		err = fmt.Errorf("unable to stat file %#v: %w", path, err)
		return
	}
	inode = stat.Ino
	size = stat.Size
	return
}

func fileWasRotatedOrTruncated(
	path string,
	ino uint64,
	size int64,
) (bool, error) {
	stat := &unix.Stat_t{}
	err := unix.Stat(path, stat)
	if err != nil {
		return false, fmt.Errorf("unable to stat file %#v: %w", path, err)
	}
	return stat.Ino != ino || stat.Size < size, nil
}
